/* 
 * Copyright (C) 2005 Remco Poelstra <remco@beryllium.net>
 *
 * LinABX - jack.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors: Remco Poelstra <remco@beryllium.net>
 */

#include <jack/jack.h>
#include <stdio.h>
#include <stdlib.h>
#include "jack.h"
#include "callbacks.h"
#include "dialog.h"
#include <strings.h>
#include <gtk/gtk.h>
#include "sample_mode.h"

jack_port_t *outputL;
jack_port_t *outputR;
jack_port_t *inputAL;
jack_port_t *inputAR;
jack_port_t *inputBL;
jack_port_t *inputBR;
jack_client_t *client;

int playing=NO_ANSWER;

enum mode_t mode=NO_MODE;

void error(const char *desc) {
    jack_error_dialog(desc);
}

int process_jack(jack_nframes_t nframes, void *arg) {
    jack_default_audio_sample_t *outL= (jack_default_audio_sample_t *) jack_port_get_buffer(outputL,nframes);
    jack_default_audio_sample_t *outR= (jack_default_audio_sample_t *) jack_port_get_buffer(outputR,nframes);
    jack_default_audio_sample_t *inL;
    jack_default_audio_sample_t *inR;
	    
    switch(playing) {
	case A:
	    inL= (jack_default_audio_sample_t *) jack_port_get_buffer(inputAL,nframes);
	    inR= (jack_default_audio_sample_t *) jack_port_get_buffer(inputAR,nframes);
	    memcpy(outL,inL,sizeof(jack_default_audio_sample_t)*nframes);
	    memcpy(outR,inR,sizeof(jack_default_audio_sample_t)*nframes);
	    break;
	case B:
	    inL= (jack_default_audio_sample_t *) jack_port_get_buffer(inputBL,nframes);
	    inR= (jack_default_audio_sample_t *) jack_port_get_buffer(inputBR,nframes);
	    memcpy(outL,inL,sizeof(jack_default_audio_sample_t)*nframes);
	    memcpy(outR,inR,sizeof(jack_default_audio_sample_t)*nframes);
	    break;
	default:
	    memset(outL,0,sizeof(jack_default_audio_sample_t)*nframes);
	    memset(outR,0,sizeof(jack_default_audio_sample_t)*nframes);
    }
    return 0;
}

int process(jack_nframes_t nframes, void *arg) {
    jack_default_audio_sample_t *outL= (jack_default_audio_sample_t *) jack_port_get_buffer(outputL,nframes);
    jack_default_audio_sample_t *outR= (jack_default_audio_sample_t *) jack_port_get_buffer(outputR,nframes);

    switch(mode) {
	case JACK_MODE:
	    return process_jack(nframes,arg);
	case SAMPLE_MODE:
	    return process_samples(nframes,arg);
	case TONE_MODE:
	    return 0;
	default:
	    memset(outL,0,sizeof(jack_default_audio_sample_t)*nframes);
	    memset(outR,0,sizeof(jack_default_audio_sample_t)*nframes);
	    return 0;
    }
}


void shutdown(void *arg) {
	exit(1);
}

int sample_rate_change(jack_nframes_t srate, void *arg) {
    update_sample_rate(srate);
    return 0;
}

int init_jack(void) {
    const char **ports;
    jack_status_t status;
    
    jack_set_error_function(error);

    while((client=jack_client_open("linabx",JackNullOption,&status))==NULL) {
	int response;
	
	response=no_server_dialog();
	if(response!=0)
	    exit(1);
    }
    jack_set_process_callback(client,process,NULL);
    jack_on_shutdown(client,shutdown,NULL);
    jack_set_sample_rate_callback(client,sample_rate_change,NULL);
    usleep(25);

    outputL=jack_port_register(client, "outputL", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput,0);
    outputR=jack_port_register(client, "outputR", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput,0);
    
    if(!outputL||!outputR) {
	failed_to_register_ports_dialog();
	jack_client_close(client);
	exit(1);
    }
    
    if (jack_activate(client)) {
	activate_failed_dialog();
	jack_client_close(client);
	exit(1);
    }
    
    if( (ports=jack_get_ports(client, NULL, NULL, JackPortIsPhysical|JackPortIsInput))==NULL) {
	no_output_ports_dialog();
	jack_client_close(client);
	exit(1);
    }

    if (jack_connect (client, jack_port_name (outputL), ports[0])||jack_connect (client, jack_port_name (outputR), ports[1])) {
	if(failed_to_connect_ports_dialog()!=GTK_RESPONSE_ACCEPT) {
	    jack_client_close(client);
	    exit(1);
	}
    }

    free(ports);

    return 0;    
}

void close_jack(void) {
    jack_client_close(client);
}

void set_playing(int answer) {
    playing=answer;
}

int init_jack_mode(void) {
    inputAL=jack_port_register(client, "inputAL", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput,0);
    inputAR=jack_port_register(client, "inputAR", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput,0);
    inputBL=jack_port_register(client, "inputBL", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput,0);
    inputBR=jack_port_register(client, "inputBR", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput,0);

    if(!inputAL||!inputAR||!inputBL||!inputBR) {
	return FALSE;
    }    
    
    mode=JACK_MODE;
    return TRUE;
}

void cleanup_jack_mode(void) {
    mode=NO_MODE;
    
    jack_port_unregister(client,inputAL);
    jack_port_unregister(client,inputAR);
    jack_port_unregister(client,inputBL);
    jack_port_unregister(client,inputBR);
    
    return;
}

const char **get_available_ports(int *A_l, int *A_r, int *B_l, int *B_r) {
    const char **ports;
    int i;
    
    *A_l=-1;
    *A_r=-1;
    *B_l=-1;
    *B_r=-1;
    ports=jack_get_ports(client, NULL, NULL, JackPortIsOutput);
    for(i=0;ports[i];i++) {
	if(jack_port_connected_to(inputAL,ports[i])) *A_l=i;
	if(jack_port_connected_to(inputAR,ports[i])) *A_r=i;
	if(jack_port_connected_to(inputBL,ports[i])) *B_l=i;
	if(jack_port_connected_to(inputBR,ports[i])) *B_r=i;
    }
    return ports;
}

void connect_port(int port,char *port_name) {
    switch(port) {
	case 0:
    	    if(!jack_port_connected_to(inputAL,port_name)) {
		jack_port_disconnect(client,inputAL);
		jack_connect(client,port_name,jack_port_name(inputAL));
	    }
	    break;
	case 1:
    	    if(!jack_port_connected_to(inputAR,port_name)) {
		jack_port_disconnect(client,inputAR);
		jack_connect(client,port_name,jack_port_name(inputAR));
	    }
	    break;
	case 2:
    	    if(!jack_port_connected_to(inputBL,port_name)) {
		jack_port_disconnect(client,inputBL);
		jack_connect(client,port_name,jack_port_name(inputBL));
	    }
	    break;
	case 3:
    	    if(!jack_port_connected_to(inputBR,port_name)) {
		jack_port_disconnect(client,inputBR);
		jack_connect(client,port_name,jack_port_name(inputBR));
	    }
	    break;
    }
}
